@@ -6,7 +6,7 @@ __all__ = [
 
 # This should always be 0.0.0 in master. Only update this after tagging
 # before release.
-__version__ = '9.2.0'
+__version__ = '10.1.1'
 
 import contextlib
 import json
@@ -17,6 +17,7 @@ import time
 import requests
 import six
 
+from datetime import datetime, timedelta
 from . import files, stone_serializers
 from .auth import (
     AuthError_validator,
@@ -50,6 +51,7 @@ from .session import (
 
 PATH_ROOT_HEADER = 'Dropbox-API-Path-Root'
 HTTP_STATUS_INVALID_PATH_ROOT = 422
+TOKEN_EXPIRATION_BUFFER = 300
 
 class RouteResult(object):
     """The successful result of a call to a route."""
@@ -126,20 +128,24 @@ class _DropboxTransport(object):
     _ROUTE_STYLE_RPC = 'rpc'
 
     # This is the default longest time we'll block on receiving data from the server
-    _DEFAULT_TIMEOUT = 30
+    _DEFAULT_TIMEOUT = 100
 
     def __init__(self,
-                 oauth2_access_token,
+                 oauth2_access_token=None,
                  max_retries_on_error=4,
                  max_retries_on_rate_limit=None,
                  user_agent=None,
                  session=None,
                  headers=None,
-                 timeout=_DEFAULT_TIMEOUT):
+                 timeout=_DEFAULT_TIMEOUT,
+                 oauth2_refresh_token=None,
+                 oauth2_access_token_expiration=None,
+                 app_key=None,
+                 app_secret=None,
+                 scope=None,):
         """
         :param str oauth2_access_token: OAuth2 access token for making client
             requests.
-
         :param int max_retries_on_error: On 5xx errors, the number of times to
             retry.
         :param Optional[int] max_retries_on_rate_limit: On 429 errors, the
@@ -158,18 +164,41 @@ class _DropboxTransport(object):
             server. After the timeout the client will give up on
             connection. If `None`, client will wait forever. Defaults
             to 30 seconds.
+        :param str oauth2_refresh_token: OAuth2 refresh token for refreshing access token
+        :param datetime oauth2_access_token_expiration: Expiration for oauth2_access_token
+        :param str app_key: application key of requesting application; used for token refresh
+        :param str app_secret: application secret of requesting application; used for token refresh
+            Not required if PKCE was used to authorize the token
+        :param list scope: list of scopes to request on refresh.  If left blank,
+            refresh will request all available scopes for application
         """
-        assert len(oauth2_access_token) > 0, \
-            'OAuth2 access token cannot be empty.'
-        assert headers is None or isinstance(headers, dict), \
-            'Expected dict, got %r' % headers
+
+        if not (oauth2_access_token or oauth2_refresh_token):
+            raise BadInputException('OAuth2 access token or refresh token must be set')
+
+        if headers is not None and not isinstance(headers, dict):
+            raise BadInputException('Expected dict, got {}'.format(headers))
+
+        if oauth2_refresh_token and not app_key:
+            raise BadInputException("app_key is required to refresh tokens")
+
+        if scope is not None and (len(scope) == 0 or not isinstance(scope, list)):
+            raise BadInputException("Scope list must be of type list")
+
         self._oauth2_access_token = oauth2_access_token
+        self._oauth2_refresh_token = oauth2_refresh_token
+        self._oauth2_access_token_expiration = oauth2_access_token_expiration
+
+        self._app_key = app_key
+        self._app_secret = app_secret
+        self._scope = scope
 
         self._max_retries_on_error = max_retries_on_error
         self._max_retries_on_rate_limit = max_retries_on_rate_limit
         if session:
-            assert isinstance(session, requests.sessions.Session), \
-                'Expected requests.sessions.Session, got %r' % session
+            if not isinstance(session, requests.sessions.Session):
+                raise BadInputException('Expected requests.sessions.Session, got {}'
+                                        .format(session))
             self._session = session
         else:
             self._session = create_session()
@@ -199,14 +228,19 @@ class _DropboxTransport(object):
             user_agent=None,
             session=None,
             headers=None,
-            timeout=None):
+            timeout=None,
+            oauth2_refresh_token=None,
+            oauth2_access_token_expiration=None,
+            app_key=None,
+            app_secret=None,
+            scope=None):
         """
         Creates a new copy of the Dropbox client with the same defaults unless modified by
         arguments to clone()
 
         See constructor for original parameter descriptions.
 
-        :return: New instance of Dropbox clent
+        :return: New instance of Dropbox client
         :rtype: Dropbox
         """
 
@@ -217,7 +251,12 @@ class _DropboxTransport(object):
             user_agent or self._user_agent,
             session or self._session,
             headers or self._headers,
-            timeout or self._timeout
+            timeout or self._timeout,
+            oauth2_refresh_token or self._oauth2_refresh_token,
+            oauth2_access_token_expiration or self._oauth2_access_token_expiration,
+            app_key or self._app_key,
+            app_secret or self._app_secret,
+            scope or self._scope
         )
 
     def request(self,
@@ -247,6 +286,9 @@ class _DropboxTransport(object):
             Dropbox object.  Defaults to `None`.
         :return: The route's result.
         """
+
+        self.check_and_refresh_access_token()
+
         host = route.attrs['host'] or 'api'
         route_name = namespace + '/' + route.name
         if route.version > 1:
@@ -299,6 +341,64 @@ class _DropboxTransport(object):
         else:
             return deserialized_result
 
+    def check_and_refresh_access_token(self):
+        """
+        Checks if access token needs to be refreshed and refreshes if possible
+        :return:
+        """
+        can_refresh = self._oauth2_refresh_token and self._app_key
+        needs_refresh = self._oauth2_access_token_expiration and \
+            (datetime.utcnow() + timedelta(seconds=TOKEN_EXPIRATION_BUFFER)) >= \
+            self._oauth2_access_token_expiration
+        needs_token = not self._oauth2_access_token
+        if (needs_refresh or needs_token) and can_refresh:
+            self.refresh_access_token(scope=self._scope)
+
+    def refresh_access_token(self, host=API_HOST, scope=None):
+        """
+        Refreshes an access token via refresh token if available
+
+        :param host: host to hit token endpoint with
+        :param scope: list of permission scopes for access token
+        :return:
+        """
+
+        if scope is not None and (len(scope) == 0 or not isinstance(scope, list)):
+            raise BadInputException("Scope list must be of type list")
+
+        if not (self._oauth2_refresh_token and self._app_key):
+            self._logger.warning('Unable to refresh access token without \
+                refresh token and app key')
+            return
+
+        self._logger.info('Refreshing access token.')
+        url = "https://{}/oauth2/token".format(host)
+        body = {'grant_type': 'refresh_token',
+                'refresh_token': self._oauth2_refresh_token,
+                'client_id': self._app_key,
+                }
+        if self._app_secret:
+            body['client_secret'] = self._app_secret
+        if scope:
+            scope = " ".join(scope)
+            body['scope'] = scope
+
+        timeout = self._DEFAULT_TIMEOUT
+        if self._timeout:
+            timeout = self._timeout
+        res = self._session.post(url, data=body, timeout=timeout)
+        if res.status_code == 400 and res.json()['error'] == 'invalid_grant':
+            request_id = res.headers.get('x-dropbox-request-id')
+            err = stone_serializers.json_compat_obj_decode(
+                AuthError_validator, 'invalid_access_token')
+            raise AuthError(request_id, err)
+        res.raise_for_status()
+
+        token_content = res.json()
+        self._oauth2_access_token = token_content["access_token"]
+        self._oauth2_access_token_expiration = datetime.utcnow() + \
+            timedelta(seconds=int(token_content["expires_in"]))
+
     def request_json_object(self,
                             host,
                             route_name,
@@ -354,6 +454,7 @@ class _DropboxTransport(object):
         """
         attempt = 0
         rate_limit_errors = 0
+        has_refreshed = False
         while True:
             self._logger.info('Request to %s', route_name)
             try:
@@ -363,6 +464,18 @@ class _DropboxTransport(object):
                                                 request_json_arg,
                                                 request_binary,
                                                 timeout=timeout)
+            except AuthError as e:
+                if e.error and e.error.is_expired_access_token():
+                    if has_refreshed:
+                        raise
+                    else:
+                        self._logger.info(
+                            'ExpiredCredentials status_code=%s: Refreshing and Retrying',
+                            e.status_code)
+                        self.refresh_access_token()
+                        has_refreshed = True
+                else:
+                    raise
             except InternalServerError as e:
                 attempt += 1
                 if attempt <= self._max_retries_on_error:
@@ -600,6 +713,8 @@ class DropboxTeam(_DropboxTransport, DropboxTeamBase):
         new_headers[select_header_name] = team_member_id
         return Dropbox(
             self._oauth2_access_token,
+            oauth2_refresh_token=self._oauth2_refresh_token,
+            oauth2_access_token_expiration=self._oauth2_access_token_expiration,
             max_retries_on_error=self._max_retries_on_error,
             max_retries_on_rate_limit=self._max_retries_on_rate_limit,
             timeout=self._timeout,
@@ -607,3 +722,12 @@ class DropboxTeam(_DropboxTransport, DropboxTeamBase):
             session=self._session,
             headers=new_headers,
         )
+
+class BadInputException(Exception):
+    """
+    Thrown if incorrect types/values are used
+
+    This should only ever be thrown during testing, app should have validation of input prior to
+    reaching this point
+    """
+    pass
