@@ -1,3 +1,5 @@
+import hashlib
+
 __all__ = [
     'BadRequestException',
     'BadStateException',
@@ -14,8 +16,14 @@ import base64
 import os
 import six
 import urllib
+import re
+from datetime import datetime, timedelta
 
-from .session import pinned_session
+from .session import (
+    API_HOST,
+    WEB_HOST,
+    pinned_session,
+)
 
 if six.PY3:
     url_path_quote = urllib.parse.quote  # pylint: disable=no-member,useless-suppression
@@ -24,31 +32,55 @@ else:
     url_path_quote = urllib.quote  # pylint: disable=no-member,useless-suppression
     url_encode = urllib.urlencode  # pylint: disable=no-member,useless-suppression
 
+TOKEN_ACCESS_TYPES = ['offline', 'online', 'legacy']
+INCLUDE_GRANTED_SCOPES_TYPES = ['user', 'team']
+PKCE_VERIFIER_LENGTH = 128
 
 class OAuth2FlowNoRedirectResult(object):
     """
     Authorization information for an OAuth2Flow performed with no redirect.
+
+    Currently refresh_token and expiration are both beta features, if you are interested
+    in using them, please contact Dropbox support
     """
 
-    def __init__(self, access_token, account_id, user_id):
+    def __init__(self, access_token, account_id, user_id, refresh_token, expiration, scope):
         """
         Args:
             access_token (str): Token to be used to authenticate later
                 requests.
-            account_id (str): The Dropbox user's account ID. Please use this
-                instead of the user_id.
-            user_id (str): For backwards compatibility with API v1, please
-                avoid using this if possible.
+            refresh_token (str): Token to be used to acquire new access token
+                when existing one expires
+            expiration (int, datetime): Either the number of seconds from now that the token expires
+                in or the datetime at which the token expires
+            account_id (str): The Dropbox user's account ID.
+            user_id (str): Deprecated (use account_id instead).
+            refresh_token (str): Token to be used to acquire new access token
+                when existing one expires
+            expiration (int, datetime): Either the number of seconds from now that the token expires
+                in or the datetime at which the token expires
+            scope (list): list of scopes to request in base oauth flow.
         """
         self.access_token = access_token
+        if not expiration:
+            self.expires_at = None
+        elif isinstance(expiration, datetime):
+            self.expires_at = expiration
+        else:
+            self.expires_at = datetime.utcnow() + timedelta(seconds=int(expiration))
+        self.refresh_token = refresh_token
         self.account_id = account_id
         self.user_id = user_id
+        self.scope = scope
 
     def __repr__(self):
-        return 'OAuth2FlowNoRedirectResult(%r, %r, %r)' % (
+        return 'OAuth2FlowNoRedirectResult(%s, %s, %s, %s, %s, %s)' % (
             self.access_token,
             self.account_id,
             self.user_id,
+            self.refresh_token,
+            self.expires_at,
+            self.scope,
         )
 
 
@@ -57,7 +89,8 @@ class OAuth2FlowResult(OAuth2FlowNoRedirectResult):
     Authorization information for an OAuth2Flow with redirect.
     """
 
-    def __init__(self, access_token, account_id, user_id, url_state):
+    def __init__(self, access_token, account_id, user_id, url_state, refresh_token,
+                 expires_in, scope):
         """
         Same as OAuth2FlowNoRedirectResult but with url_state.
 
@@ -66,51 +99,97 @@ class OAuth2FlowResult(OAuth2FlowNoRedirectResult):
                 :meth:`DropboxOAuth2Flow.start`.
         """
         super(OAuth2FlowResult, self).__init__(
-            access_token, account_id, user_id)
+            access_token=access_token,
+            account_id=account_id,
+            user_id=user_id,
+            refresh_token=refresh_token,
+            expires_in=expires_in,
+            scope=scope)
         self.url_state = url_state
 
     @classmethod
     def from_no_redirect_result(cls, result, url_state):
         assert isinstance(result, OAuth2FlowNoRedirectResult)
-        return cls(
-            result.access_token, result.account_id, result.user_id, url_state)
+        return cls(result.access_token, result.account_id, result.user_id,
+                   url_state, result.refresh_token, result.expires_at, result.scope)
 
     def __repr__(self):
-        return 'OAuth2FlowResult(%r, %r, %r, %r)' % (
+        return 'OAuth2FlowResult(%s, %s, %s, %s, %s, %s, %s, %s, %s)' % (
             self.access_token,
+            self.refresh_token,
+            self.expires_at,
             self.account_id,
             self.user_id,
+            self.scope,
             self.url_state,
+            self.refresh_token,
+            self.expires_at,
         )
 
 
 class DropboxOAuth2FlowBase(object):
 
-    def __init__(self, consumer_key, consumer_secret, locale=None):
+    def __init__(self, consumer_key, consumer_secret=None, locale=None, token_access_type='legacy',
+                 scope=None, include_granted_scopes=None, use_pkce=False):
+        if scope is not None and (len(scope) == 0 or not isinstance(scope, list)):
+            raise BadInputException("Scope list must be of type list")
+        if token_access_type is not None and token_access_type not in TOKEN_ACCESS_TYPES:
+            raise BadInputException("Token access type must be from the following enum: {}".format(
+                TOKEN_ACCESS_TYPES))
+        if not (use_pkce or consumer_secret):
+            raise BadInputException("Must pass in either consumer secret or use PKCE")
+        if include_granted_scopes and not scope:
+            raise BadInputException("Must pass in scope to pass include_granted_scopes")
+
         self.consumer_key = consumer_key
         self.consumer_secret = consumer_secret
         self.locale = locale
+        self.token_access_type = token_access_type
         self.requests_session = pinned_session()
+        self.scope = scope
+        self.include_granted_scopes = include_granted_scopes
 
-        self._host = os.environ.get('DROPBOX_WEB_HOST', 'www.dropbox.com')
+        if use_pkce:
+            self.code_verifier = _generate_pkce_code_verifier()
+            self.code_challenge = _generate_pkce_code_challenge(self.code_verifier)
+        else:
+            self.code_verifier = None
+            self.code_challenge = None
 
-    def _get_authorize_url(self, redirect_uri, state):
+    def _get_authorize_url(self, redirect_uri, state, token_access_type, scope=None,
+                           include_granted_scopes=None, code_challenge=None):
         params = dict(response_type='code',
                       client_id=self.consumer_key)
         if redirect_uri is not None:
             params['redirect_uri'] = redirect_uri
         if state is not None:
             params['state'] = state
+        if token_access_type is not None:
+            assert token_access_type in TOKEN_ACCESS_TYPES
+            if token_access_type != 'legacy':
+                params['token_access_type'] = token_access_type
+        if code_challenge:
+            params['code_challenge'] = code_challenge
+            params['code_challenge_method'] = 'S256'
 
-        return self.build_url('/oauth2/authorize', params)
+        if scope is not None:
+            params['scope'] = " ".join(scope)
+            if include_granted_scopes is not None:
+                assert include_granted_scopes in INCLUDE_GRANTED_SCOPES_TYPES
+                params['include_granted_scopes'] = include_granted_scopes
 
-    def _finish(self, code, redirect_uri):
+        return self.build_url('/oauth2/authorize', params, WEB_HOST)
+
+    def _finish(self, code, redirect_uri, code_verifier):
         url = self.build_url('/oauth2/token')
         params = {'grant_type': 'authorization_code',
                   'code': code,
                   'client_id': self.consumer_key,
-                  'client_secret': self.consumer_secret,
                   }
+        if code_verifier:
+            params['code_verifier'] = code_verifier
+        else:
+            params['client_secret'] = self.consumer_secret
         if self.locale is not None:
             params['locale'] = self.locale
         if redirect_uri is not None:
@@ -121,10 +200,37 @@ class DropboxOAuth2FlowBase(object):
 
         d = resp.json()
 
+        if 'team_id' in d:
+            account_id = d['team_id']
+        else:
+            account_id = d['account_id']
+
+        access_token = d['access_token']
+
+        if 'refresh_token' in d:
+            refresh_token = d['refresh_token']
+        else:
+            refresh_token = ""
+
+        if 'expires_in' in d:
+            expires_in = d['expires_in']
+        else:
+            expires_in = None
+
+        if 'scope' in d:
+            scope = d['scope']
+        else:
+            scope = None
+
+        uid = d['uid']
+
         return OAuth2FlowNoRedirectResult(
-            d['access_token'],
-            d['account_id'],
-            d['uid'])
+            access_token,
+            account_id,
+            uid,
+            refresh_token,
+            expires_in,
+            scope)
 
     def build_path(self, target, params=None):
         """Build the path component for an API URL.
@@ -155,7 +261,7 @@ class DropboxOAuth2FlowBase(object):
         else:
             return target_path
 
-    def build_url(self, target, params=None):
+    def build_url(self, target, params=None, host=API_HOST):
         """Build an API URL.
 
         This method adds scheme and hostname to the path
@@ -166,7 +272,7 @@ class DropboxOAuth2FlowBase(object):
         :return: The full API URL.
         :rtype: str
         """
-        return "https://%s%s" % (self._host, self.build_path(target, params))
+        return "https://%s%s" % (host, self.build_path(target, params))
 
 
 class DropboxOAuth2FlowNoRedirect(DropboxOAuth2FlowBase):
@@ -181,21 +287,22 @@ class DropboxOAuth2FlowNoRedirect(DropboxOAuth2FlowBase):
         auth_flow = DropboxOAuth2FlowNoRedirect(APP_KEY, APP_SECRET)
 
         authorize_url = auth_flow.start()
-        print "1. Go to: " + authorize_url
-        print "2. Click \\"Allow\\" (you might have to log in first)."
-        print "3. Copy the authorization code."
+        print("1. Go to: " + authorize_url)
+        print("2. Click \\"Allow\\" (you might have to log in first).")
+        print("3. Copy the authorization code.")
         auth_code = raw_input("Enter the authorization code here: ").strip()
 
         try:
             oauth_result = auth_flow.finish(auth_code)
-        except Exception, e:
+        except Exception as e:
             print('Error: %s' % (e,))
             return
 
         dbx = Dropbox(oauth_result.access_token)
     """
 
-    def __init__(self, consumer_key, consumer_secret, locale=None):
+    def __init__(self, consumer_key, consumer_secret=None, locale=None, token_access_type='legacy',
+                 scope=None, include_granted_scopes=None, use_pkce=False):  # noqa: E501;
         """
         Construct an instance.
 
@@ -206,12 +313,30 @@ class DropboxOAuth2FlowNoRedirect(DropboxOAuth2FlowBase):
             example "en" or "en_US". Some API calls return localized data and
             error messages; this setting tells the server which locale to use.
             By default, the server uses "en_US".
+        :param str token_access_type: the type of token to be requested.
+            From the following enum:
+            legacy - creates one long-lived token with no expiration
+            online - create one short-lived token with an expiration
+            offline - create one short-lived token with an expiration with a refresh token
+        :param list scope: list of scopes to request in base oauth flow.  If left blank,
+            will default to all scopes for app
+        :param str include_granted_scopes: which scopes to include from previous grants
+            From the following enum:
+            user - include user scopes in the grant
+            team - include team scopes in the grant
+            Note: if this user has never linked the app, include_granted_scopes must be None
+        :param bool use_pkce: Whether or not to use Sha256 based PKCE. PKCE should be only use on
+            client apps which doesn't call your server. It is less secure than non-PKCE flow but
+            can be used if you are unable to safely retrieve your app secret
         """
-        # pylint: disable=useless-super-delegation
         super(DropboxOAuth2FlowNoRedirect, self).__init__(
-            consumer_key,
-            consumer_secret,
-            locale,
+            consumer_key=consumer_key,
+            consumer_secret=consumer_secret,
+            locale=locale,
+            token_access_type=token_access_type,
+            scope=scope,
+            include_granted_scopes=include_granted_scopes,
+            use_pkce=use_pkce,
         )
 
     def start(self):
@@ -223,7 +348,10 @@ class DropboxOAuth2FlowNoRedirect(DropboxOAuth2FlowBase):
             access the user's Dropbox account. Tell the user to visit this URL
             and approve your app.
         """
-        return self._get_authorize_url(None, None)
+        return self._get_authorize_url(None, None, self.token_access_type,
+                                       scope=self.scope,
+                                       include_granted_scopes=self.include_granted_scopes,
+                                       code_challenge=self.code_challenge)
 
     def finish(self, code):
         """
@@ -236,7 +364,7 @@ class DropboxOAuth2FlowNoRedirect(DropboxOAuth2FlowBase):
         :rtype: OAuth2FlowNoRedirectResult
         :raises: The same exceptions as :meth:`DropboxOAuth2Flow.finish()`.
         """
-        return self._finish(code, None)
+        return self._finish(code, None, self.code_verifier)
 
 
 class DropboxOAuth2Flow(DropboxOAuth2FlowBase):
@@ -284,8 +412,10 @@ class DropboxOAuth2Flow(DropboxOAuth2FlowBase):
 
     """
 
-    def __init__(self, consumer_key, consumer_secret, redirect_uri, session,
-                 csrf_token_session_key, locale=None):
+    def __init__(self, consumer_key, redirect_uri, session,
+                 csrf_token_session_key, consumer_secret=None, locale=None,
+                 token_access_type='legacy', scope=None,
+                include_granted_scopes=None, use_pkce=False):
         """
         Construct an instance.
 
@@ -304,8 +434,28 @@ class DropboxOAuth2Flow(DropboxOAuth2FlowBase):
             example "en" or "en_US". Some API calls return localized data and
             error messages; this setting tells the server which locale to use.
             By default, the server uses "en_US".
+        :param str token_access_type: the type of token to be requested.
+            From the following enum:
+            legacy - creates one long-lived token with no expiration
+            online - create one short-lived token with an expiration
+            offline - create one short-lived token with an expiration with a refresh token
+        :param list scope: list of scopes to request in base oauth flow.  If left blank,
+            will default to all scopes for app
+        :param str include_granted_scopes: which scopes to include from previous grants
+            From the following enum:
+            user - include user scopes in the grant
+            team - include team scopes in the grant
+            Note: if this user has never linked the app, include_granted_scopes must be None
+        :param bool use_pkce: Whether or not to use Sha256 based PKCE
         """
-        super(DropboxOAuth2Flow, self).__init__(consumer_key, consumer_secret, locale)
+        super(DropboxOAuth2Flow, self).__init__(
+            consumer_key=consumer_key,
+            consumer_secret=consumer_secret,
+            locale=locale,
+            token_access_type=token_access_type,
+            scope=scope,
+            include_granted_scopes=include_granted_scopes,
+            use_pkce=use_pkce)
         self.redirect_uri = redirect_uri
         self.session = session
         self.csrf_token_session_key = csrf_token_session_key
@@ -339,7 +489,10 @@ class DropboxOAuth2Flow(DropboxOAuth2FlowBase):
             state += "|" + url_state
         self.session[self.csrf_token_session_key] = csrf_token
 
-        return self._get_authorize_url(self.redirect_uri, state)
+        return self._get_authorize_url(self.redirect_uri, state, self.token_access_type,
+                                       scope=self.scope,
+                                       include_granted_scopes=self.include_granted_scopes,
+                                       code_challenge=self.code_challenge)
 
     def finish(self, query_params):
         """
@@ -423,7 +576,7 @@ class DropboxOAuth2Flow(DropboxOAuth2FlowBase):
 
         # If everything went ok, make the network call to get an access token.
 
-        no_redirect_result = self._finish(code, self.redirect_uri)
+        no_redirect_result = self._finish(code, self.redirect_uri, self.code_verifier)
         return OAuth2FlowResult.from_no_redirect_result(
             no_redirect_result, url_state)
 
@@ -477,6 +630,16 @@ class ProviderException(Exception):
     pass
 
 
+class BadInputException(Exception):
+    """
+    Thrown if incorrect types/values are used
+
+    This should only ever be thrown during testing, app should have validation of input prior to
+    reaching this point
+    """
+    pass
+
+
 def _safe_equals(a, b):
     if len(a) != len(b):
         return False
@@ -505,3 +668,16 @@ def _params_to_urlencoded(params):
 
     utf8_params = {encode(k): encode(v) for k, v in six.iteritems(params)}
     return url_encode(utf8_params)
+
+def _generate_pkce_code_verifier():
+    code_verifier = base64.urlsafe_b64encode(os.urandom(PKCE_VERIFIER_LENGTH)).decode('utf-8')
+    code_verifier = re.sub('[^a-zA-Z0-9]+', '', code_verifier)
+    if len(code_verifier) > PKCE_VERIFIER_LENGTH:
+        code_verifier = code_verifier[:128]
+    return code_verifier
+
+def _generate_pkce_code_challenge(code_verifier):
+    code_challenge = hashlib.sha256(code_verifier.encode('utf-8')).digest()
+    code_challenge = base64.urlsafe_b64encode(code_challenge).decode('utf-8')
+    code_challenge = code_challenge.replace('=', '')
+    return code_challenge
